/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-BSD-5-Clause-Nordic
 */

#include <zephyr.h>
#include <stdio.h>
#include <nrf_socket.h>
#include <net/socket.h>
#include <drivers/uart.h>
#include <string.h>
#include <dk_buttons_and_leds.h>

#include <net/mqtt.h>
#include <net/socket.h>
#include <modem/lte_lc.h>
#if defined(CONFIG_LWM2M_CARRIER)
#include <lwm2m_carrier.h>
#endif

#define AT_XSYSTEMMODE    "AT\%XSYSTEMMODE=1,0,1,0"
#define AT_ACTIVATE_GPS_LTE_SAME   "AT+CFUN=31"
#define AT_ACTIVATE_LTE_GPS_SAME   "AT+CFUN=21"
#define AT_DEACTIVATE_GPS_LTE_SAME   "AT+CFUN=30"
#define AT_DEACTIVATE_LTE_GPS_SAME   "AT+CFUN=20"
#define AT_DEACTIVATE_LTE "AT+CFUN=20"

#define AT_CMD_SIZE(x) (sizeof(x) - 1)


//#ifdef CONFIG_BOARD_NRF9160_PCA10090NS
#define AT_MAGPIO      "AT\%XMAGPIO=1,0,0,1,1,1574,1577"
#define AT_COEX0       "AT\%XCOEX0=1,1,1570,1580"
//#endif

static const char     update_indicator[] = {'\\', '|', '/', '-'};
static const char     at_commands[][31]  = {
				AT_XSYSTEMMODE,
#ifdef CONFIG_BOARD_NRF9160_PCA10090NS
				AT_MAGPIO,
				AT_COEX0,
#endif
				AT_ACTIVATE_GPS_LTE_SAME
			};

static int            gnss_fd = -1;
static int            connect_count = 0;
static char           nmea_strings[10][NRF_GNSS_NMEA_MAX_LEN];
static u32_t          nmea_string_cnt;

static bool           got_first_fix;
static bool           update_terminal;
static bool           LTEMode = true;
static bool           gps_active = false;
static bool           lte_active = false;
static u64_t          fix_timestamp;
nrf_gnss_data_frame_t last_fix;

K_SEM_DEFINE(lte_ready, 0, 1);

/* Buffers for MQTT client. */
static u8_t rx_buffer[CONFIG_MQTT_MESSAGE_BUFFER_SIZE];
static u8_t tx_buffer[CONFIG_MQTT_MESSAGE_BUFFER_SIZE];
static u8_t payload_buf[CONFIG_MQTT_PAYLOAD_BUFFER_SIZE];

/* The mqtt client struct */
static struct mqtt_client client;

/* MQTT Broker details. */
static struct sockaddr_storage broker;

/* Connected flag */
static bool connected;

/* File descriptor */
static struct pollfd fds;
 
#if defined(CONFIG_MQTT_LIB_TLS)
static sec_tag_t sec_tag_list[] = { CONFIG_SEC_TAG };
#endif /* defined(CONFIG_MQTT_LIB_TLS) */ 


#if defined(CONFIG_BSD_LIBRARY)

/**@brief Recoverable BSD library error. */
void bsd_recoverable_error_handler(uint32_t err)
{
	printk("bsdlib recoverable error: %u\n", (unsigned int)err);
}

#endif /* defined(CONFIG_BSD_LIBRARY) */

#if defined(CONFIG_LWM2M_CARRIER)
K_SEM_DEFINE(carrier_registered, 0, 1);

void lwm2m_carrier_event_handler(const lwm2m_carrier_event_t *event)
{
	switch (event->type) {
	case LWM2M_CARRIER_EVENT_BSDLIB_INIT:
		printk("LWM2M_CARRIER_EVENT_BSDLIB_INIT\n");
		break;
	case LWM2M_CARRIER_EVENT_CONNECT:
		printk("LWM2M_CARRIER_EVENT_CONNECT\n");
		break;
	case LWM2M_CARRIER_EVENT_DISCONNECT:
		printk("LWM2M_CARRIER_EVENT_DISCONNECT\n");
		break;
	case LWM2M_CARRIER_EVENT_READY:
		printk("LWM2M_CARRIER_EVENT_READY\n");
		k_sem_give(&carrier_registered);
		break;
	case LWM2M_CARRIER_EVENT_FOTA_START:
		printk("LWM2M_CARRIER_EVENT_FOTA_START\n");
		break;
	case LWM2M_CARRIER_EVENT_REBOOT:
		printk("LWM2M_CARRIER_EVENT_REBOOT\n");
		break;
	}
}
#endif /* defined(CONFIG_LWM2M_CARRIER) */

/**@brief Function to print strings without null-termination
 */
static void data_print(u8_t *prefix, u8_t *data, size_t len)
{
	char buf[len + 1];

	memcpy(buf, data, len);
	buf[len] = 0;
	printk("%s%s\n", prefix, buf);
}

/**@brief Function to publish data on the configured topic
 */
static int data_publish(struct mqtt_client *c, enum mqtt_qos qos,
	u8_t *data, size_t len)
{
	struct mqtt_publish_param param;

	param.message.topic.qos = qos;
	param.message.topic.topic.utf8 = CONFIG_MQTT_PUB_TOPIC;
	param.message.topic.topic.size = strlen(CONFIG_MQTT_PUB_TOPIC);
	param.message.payload.data = data;
	param.message.payload.len = len;
	param.message_id = sys_rand32_get();
	param.dup_flag = 0;
	param.retain_flag = 0;

	data_print("Publishing: ", data, len);
	printk("to topic: %s len: %u\n",
		CONFIG_MQTT_PUB_TOPIC,
		(unsigned int)strlen(CONFIG_MQTT_PUB_TOPIC));

	return mqtt_publish(c, &param);
}

/**@brief Function to unsubscribe to the configured topic
 */
static int unsubscribe(void)
{
	struct mqtt_topic subscribe_topic = {
		.topic = {
			.utf8 = CONFIG_MQTT_SUB_TOPIC,
			.size = strlen(CONFIG_MQTT_SUB_TOPIC)
		},
		.qos = MQTT_QOS_1_AT_LEAST_ONCE
	};

	const struct mqtt_subscription_list subscription_list = {
		.list = &subscribe_topic,
		.list_count = 1,
		.message_id = 1234
	};

	printk("Unsubscribing from: %s len %u\n", CONFIG_MQTT_SUB_TOPIC,
		(unsigned int)strlen(CONFIG_MQTT_SUB_TOPIC));

	return mqtt_unsubscribe(&client, &subscription_list);
}

/**@brief Function to subscribe to the configured topic
 */
static int subscribe(void)
{
	struct mqtt_topic subscribe_topic = {
		.topic = {
			.utf8 = CONFIG_MQTT_SUB_TOPIC,
			.size = strlen(CONFIG_MQTT_SUB_TOPIC)
		},
		.qos = MQTT_QOS_1_AT_LEAST_ONCE
	};

	const struct mqtt_subscription_list subscription_list = {
		.list = &subscribe_topic,
		.list_count = 1,
		.message_id = 1234
	};

	printk("Subscribing to: %s len %u\n", CONFIG_MQTT_SUB_TOPIC,
		(unsigned int)strlen(CONFIG_MQTT_SUB_TOPIC));

	return mqtt_subscribe(&client, &subscription_list);
}

/**@brief Function to read the published payload.
 */
static int publish_get_payload(struct mqtt_client *c, size_t length)
{
	u8_t *buf = payload_buf;
	u8_t *end = buf + length;

	if (length > sizeof(payload_buf)) {
		return -EMSGSIZE;
	}

	while (buf < end) {
		int ret = mqtt_read_publish_payload(c, buf, end - buf);

		if (ret < 0) {
			int err;

			if (ret != -EAGAIN) {
				return ret;
			}

			printk("mqtt_read_publish_payload: EAGAIN\n");


			err = poll(&fds, 1, CONFIG_MQTT_KEEPALIVE * MSEC_PER_SEC);
			if (err > 0 && (fds.revents & POLLIN) == POLLIN) {
				continue;
			} else {
				return -EIO;
			}
		}

		if (ret == 0) {
			return -EIO;
		}

		buf += ret;
	}

	return 0;
}

/**@brief MQTT client event handler
 */
void mqtt_evt_handler(struct mqtt_client *const c,
		      const struct mqtt_evt *evt)
{
	int err;

	switch (evt->type) {
	case MQTT_EVT_CONNACK:
		if (evt->result != 0) {
			printk("MQTT connect failed %d\n", evt->result);
			break;
		}

		connected = true;
		printk("[%s:%d] MQTT client connected!\n", __func__, __LINE__);
		subscribe();
		break;

	case MQTT_EVT_DISCONNECT:
		printk("[%s:%d] MQTT client disconnected %d\n", __func__,
		       __LINE__, evt->result);

		connected = false;
		break;

	case MQTT_EVT_PUBLISH: {
		const struct mqtt_publish_param *p = &evt->param.publish;

		printk("[%s:%d] MQTT PUBLISH result=%d len=%d\n", __func__,
		       __LINE__, evt->result, p->message.payload.len);
		err = publish_get_payload(c, p->message.payload.len);
		if (err >= 0) {
			data_print("Received: ", payload_buf,
				p->message.payload.len);
			/* Echo back received data */
			data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE,
				payload_buf, p->message.payload.len);
            printk(payload_buf);
            printk("\nlength: %d\n", p->message.payload.len);
		} else {
			printk("mqtt_read_publish_payload: Failed! %d\n", err);
			printk("Disconnecting MQTT client...\n");

			err = mqtt_disconnect(c);
			if (err) {
				printk("Could not disconnect: %d\n", err);
			}
		}
	} break;

	case MQTT_EVT_PUBACK:
		if (evt->result != 0) {
			printk("MQTT PUBACK error %d\n", evt->result);
			break;
		}

		printk("[%s:%d] PUBACK packet id: %u\n", __func__, __LINE__,
				evt->param.puback.message_id);
		break;

	case MQTT_EVT_SUBACK:
		if (evt->result != 0) {
			printk("MQTT SUBACK error %d\n", evt->result);
			break;
		}

		printk("[%s:%d] SUBACK packet id: %u\n", __func__, __LINE__,
				evt->param.suback.message_id);
		break;

	default:
		printk("[%s:%d] default: %d\n", __func__, __LINE__,
				evt->type);
		break;
	}
}

/**@brief Resolves the configured hostname and
 * initializes the MQTT broker structure
 */
static void broker_init(void)
{
	int err;
	struct addrinfo *result;
	struct addrinfo *addr;
	struct addrinfo hints = {
		.ai_family = AF_INET,
		.ai_socktype = SOCK_STREAM
	};

	err = getaddrinfo(CONFIG_MQTT_BROKER_HOSTNAME, NULL, &hints, &result);
	if (err) {
		printk("ERROR: getaddrinfo failed %d\n", err);

		return;
	}

	addr = result;
	err = -ENOENT;

	/* Look for address of the broker. */
	while (addr != NULL) {
		/* IPv4 Address. */
		if (addr->ai_addrlen == sizeof(struct sockaddr_in)) {
			struct sockaddr_in *broker4 =
				((struct sockaddr_in *)&broker);
			char ipv4_addr[NET_IPV4_ADDR_LEN];

			broker4->sin_addr.s_addr =
				((struct sockaddr_in *)addr->ai_addr)
				->sin_addr.s_addr;
			broker4->sin_family = AF_INET;
			broker4->sin_port = htons(CONFIG_MQTT_BROKER_PORT);

			inet_ntop(AF_INET, &broker4->sin_addr.s_addr,
				  ipv4_addr, sizeof(ipv4_addr));
			printk("IPv4 Address found %s\n", ipv4_addr);

			break;
		} else {
			printk("ai_addrlen = %u should be %u or %u\n",
				(unsigned int)addr->ai_addrlen,
				(unsigned int)sizeof(struct sockaddr_in),
				(unsigned int)sizeof(struct sockaddr_in6));
		}

		addr = addr->ai_next;
		break;
	}
	/* Free the address. */
	freeaddrinfo(result);
}

/**@brief Initialize the MQTT client structure
 */
static void client_init(struct mqtt_client *client)
{
	mqtt_client_init(client);

	/* MQTT client configuration */
	client->broker = &broker;
	client->evt_cb = mqtt_evt_handler;
	client->client_id.utf8 = (u8_t *)CONFIG_MQTT_CLIENT_ID;
	client->client_id.size = strlen(CONFIG_MQTT_CLIENT_ID);
	client->password = NULL;
	client->user_name = NULL;
	client->protocol_version = MQTT_VERSION_3_1_1;

	/* MQTT buffers configuration */
	client->rx_buf = rx_buffer;
	client->rx_buf_size = sizeof(rx_buffer);
	client->tx_buf = tx_buffer;
	client->tx_buf_size = sizeof(tx_buffer);

#if defined(CONFIG_MQTT_LIB_TLS)
    struct mqtt_sec_config *tls_config = &client->transport.tls.config;
    
    client->transport.type = MQTT_TRANSPORT_SECURE;
    
    tls_config->peer_verify = CONFIG_PEER_VERIFY;
    tls_config->cipher_count = 0;
    tls_config->cipher_list = NULL;
    tls_config->sec_tag_count = ARRAY_SIZE(sec_tag_list);
    tls_config->sec_tag_list = sec_tag_list;
    tls_config->hostname = CONFIG_MQTT_BROKER_HOSTNAME;
#else /* MQTT transport configuration */
    client->transport.type = MQTT_TRANSPORT_NON_SECURE;
#endif /* defined(CONFIG_MQTT_LIB_TLS) */
}

/**@brief Initialize the file descriptor structure used by poll.
 */
static int fds_init(struct mqtt_client *c)
{
	if (c->transport.type == MQTT_TRANSPORT_NON_SECURE) {
		fds.fd = c->transport.tcp.sock;
	} else {
#if defined(CONFIG_MQTT_LIB_TLS)
		fds.fd = c->transport.tls.sock;
#else
		return -ENOTSUP;
#endif
	}

	fds.events = POLLIN;

	return 0;
}

/**@brief Configures modem to provide LTE link. Blocks until link is
 * successfully established.
 */
static void modem_configure(void)
{
        // activate GPS if needed
        if (gps_active) {
          if (at_cmd_write(AT_DEACTIVATE_GPS_LTE_SAME, NULL, 0, NULL) != 0) {
                  return -1;
          }
          gps_active = false;
        }
#if defined(CONFIG_LTE_LINK_CONTROL)
	if (IS_ENABLED(CONFIG_LTE_AUTO_INIT_AND_CONNECT)) {
		/* Do nothing, modem is already turned on
		 * and connected.
		 */
	} else {
#if defined(CONFIG_LWM2M_CARRIER)
		/* Wait for the LWM2M_CARRIER to configure the modem and
		 * start the connection.
		 */
		printk("Waitng for carrier registration...\n");
		k_sem_take(&carrier_registered, K_FOREVER);
		printk("Registered!\n");
#else /* defined(CONFIG_LWM2M_CARRIER) */
		int err;
                // set APN to rogers
//                if (at_cmd_write("AT+CGDCONT=1,\"IP\",\"m2minternet.apn\"", NULL, 0, NULL) != 0) {
//                        return -1;
//                }
//                if (at_cmd_write("AT+CGDCONT=1,\"IP\",\"ciot\"", NULL, 0, NULL) != 0) {
//                        return -1;
//                }
		printk("LTE Link Connecting ...\n");

		err = lte_lc_init_and_connect();
                // this sets CFUN=1, which enables GPS so we need to reflect that in the flag
                gps_active = true;
                lte_active = true;
		__ASSERT(err == 0, "LTE link could not be established.");
//		err = lte_lc_psm_req(true);
//                printk("main::modem_configure: lte_lc_psm_req err = %d\n", err);
//		if (err) {
//			printk("lte_lc_psm_req, error: %d\n", err);
//		}
		printk("LTE Link Connected!\n");
#endif /* defined(CONFIG_LWM2M_CARRIER) */
	}
#endif /* defined(CONFIG_LTE_LINK_CONTROL) */
}

static int setup_modem(void)
{
	for (int i = 0; i < ARRAY_SIZE(at_commands); i++) {
//		if (at_cmd_write(at_commands[i], NULL, 0, NULL) != 0) {
//			return -1;
//		}
            printk(at_commands[i]);
            printk("\n");
	}

	return 0;
}

static int init_gps(void)
{
	int retval;

	nrf_gnss_fix_retry_t    fix_retry    = 0;
	nrf_gnss_fix_interval_t fix_interval = 1;
	nrf_gnss_delete_mask_t	delete_mask  = 0;
	nrf_gnss_nmea_mask_t	nmea_mask    = NRF_GNSS_NMEA_GSV_MASK |
					       NRF_GNSS_NMEA_GSA_MASK |
					       NRF_GNSS_NMEA_GLL_MASK |
					       NRF_GNSS_NMEA_GGA_MASK |
                                               NRF_GNSS_NMEA_RMC_MASK;

//	if (setup_modem() != 0) {
//		printk("Failed to initialize modem\n");
//		return -1;
//	}

//        if (at_cmd_write(AT_XSYSTEMMODE, NULL, 0, NULL) != 0) {
//                return -1;
//        }
        // needed for nrf9160DK configuration
        if (at_cmd_write(AT_MAGPIO, NULL, 0, NULL) != 0) {
                return -1;
        }
        if (at_cmd_write(AT_COEX0, NULL, 0, NULL) != 0) {
                return -1;
        }
        // activate GPS if needed
        if (!gps_active) {
          if (at_cmd_write(AT_ACTIVATE_GPS_LTE_SAME, NULL, 0, NULL) != 0) {
                  return -1;
          }
          gps_active = true;
        }

	gnss_fd = nrf_socket(NRF_AF_LOCAL, NRF_SOCK_DGRAM, NRF_PROTO_GNSS);

	if (gnss_fd >= 0) {
		printk("Socket created\n");
	} else {
		printk("Could not init socket (err: %d)\n", gnss_fd);
		return -1;
	}

	retval = nrf_setsockopt(gnss_fd,
				NRF_SOL_GNSS,
				NRF_SO_GNSS_FIX_RETRY,
				&fix_retry,
				sizeof(fix_retry));

	if (retval != 0) {
		printk("Failed to set fix retry value\n");
		return -1;
	}

	retval = nrf_setsockopt(gnss_fd,
				NRF_SOL_GNSS,
				NRF_SO_GNSS_FIX_INTERVAL,
				&fix_interval,
				sizeof(fix_interval));

	if (retval != 0) {
		printk("Failed to set fix interval value\n");
		return -1;
	}

	retval = nrf_setsockopt(gnss_fd,
				NRF_SOL_GNSS,
				NRF_SO_GNSS_NMEA_MASK,
				&nmea_mask,
				sizeof(nmea_mask));

	if (retval != 0) {
		printk("Failed to set nmea mask\n");
		return -1;
	}

	retval = nrf_setsockopt(gnss_fd,
				NRF_SOL_GNSS,
				NRF_SO_GNSS_START,
				&delete_mask,
				sizeof(delete_mask));

	if (retval != 0) {
		printk("Failed to start GPS\n");
		return -1;
	}

	return 0;
}

static void print_satellite_stats(nrf_gnss_data_frame_t *pvt_data)
{
	u8_t  tracked          = 0;
	u8_t  in_fix           = 0;
	u8_t  unhealthy        = 0;

	for (int i = 0; i < NRF_GNSS_MAX_SATELLITES; ++i) {

		if ((pvt_data->pvt.sv[i].sv > 0) &&
		    (pvt_data->pvt.sv[i].sv < 33)) {

			tracked++;

			if (pvt_data->pvt.sv[i].flags &
					NRF_GNSS_SV_FLAG_USED_IN_FIX) {
				in_fix++;
			}

			if (pvt_data->pvt.sv[i].flags &
					NRF_GNSS_SV_FLAG_UNHEALTHY) {
				unhealthy++;
			}
		}
	}

	printk("Tracking: %d Using: %d Unhealthy: %d", tracked,
						       in_fix,
						       unhealthy);

	printk("\nSeconds since last fix %lld\n",
			(k_uptime_get() - fix_timestamp) / 1000);
}

static void print_pvt_data(nrf_gnss_data_frame_t *pvt_data)
{
	printf("Longitude:  %f\n", pvt_data->pvt.longitude);
	printf("Latitude:   %f\n", pvt_data->pvt.latitude);
	printf("Altitude:   %f\n", pvt_data->pvt.altitude);
	printf("Speed:      %f\n", pvt_data->pvt.speed);
	printf("Heading:    %f\n", pvt_data->pvt.heading);
	printk("Date:       %02u-%02u-%02u\n", pvt_data->pvt.datetime.day,
					       pvt_data->pvt.datetime.month,
					       pvt_data->pvt.datetime.year);
	printk("Time (UTC): %02u:%02u:%02u\n", pvt_data->pvt.datetime.hour,
					       pvt_data->pvt.datetime.minute,
					      pvt_data->pvt.datetime.seconds);
}

static void print_nmea_data(void)
{
	printk("NMEA strings:\n");

	for (int i = 0; i < nmea_string_cnt; ++i) {
		printk("%s\n", nmea_strings[i]);
	}
}

int process_gps_data(nrf_gnss_data_frame_t *gps_data)
{
	int retval;

	retval = nrf_recv(gnss_fd,
			  gps_data,
			  sizeof(nrf_gnss_data_frame_t),
			  NRF_MSG_DONTWAIT);

	if (retval > 0) {

		switch (gps_data->data_id) {
		case NRF_GNSS_PVT_DATA_ID:

			if ((gps_data->pvt.flags &
				NRF_GNSS_PVT_FLAG_FIX_VALID_BIT)
				== NRF_GNSS_PVT_FLAG_FIX_VALID_BIT) {

				if (!got_first_fix) {
					got_first_fix = true;
				}

				fix_timestamp = k_uptime_get();
				memcpy(&last_fix,
				       gps_data,
				       sizeof(nrf_gnss_data_frame_t));

				nmea_string_cnt = 0;
				update_terminal = true;
			}
			break;

		case NRF_GNSS_NMEA_DATA_ID:
			if (nmea_string_cnt < 10) {
				memcpy(nmea_strings[nmea_string_cnt++],
				       gps_data->nmea,
				       retval);
			}
			break;

		case NRF_GNSS_AGPS_DATA_ID:
			break;

		default:
			break;
		}
	}

	return retval;
}

int8_t poll_and_check_rcv_data(u32_t interval_msec)
{
	int8_t err = poll(&fds, 1, interval_msec); // Check if any data comes through LTE-M. When it is waiting for an incoming data, it sleeps
	if (err < 0) {
		printk("ERROR: poll %d\n", errno);
		return -1;
	}

	/* Check if data is received through LTE-M */
	if ((fds.revents & POLLIN) == POLLIN) { // If get data
		err = mqtt_input(&client);
		if (err < 0) {
			printk("ERROR: poll %d\n", errno);
			return -1;
		}
	}
	if ((fds.revents & POLLERR) == POLLERR) {
		printk("POLLERR\n");
		return -1;
	}
	if ((fds.revents & POLLNVAL) == POLLNVAL) {
		printk("POLLNVAL\n");
		return -1;
	}

	return 0;
}

int8_t mqtt_init()
{
        int err = 0;

        err = mqtt_connect(&client);
	if (err != 0) {
		printk("ERROR: mqtt_connect %d\n", err);
		return -1;
	}

	err = fds_init(&client);
	if (err != 0) {
		printk("ERROR: fds_init %d\n", err);
		return -1;
	}

        err = poll_and_check_rcv_data(10 * MSEC_PER_SEC); // Wait for CONNACK
        if (err != 0) {
                printk("ERROR: poll_and_check_rcv_data %d\n", err);
                return -1;
        }
}

int8_t mqtt_uninit()
{
        int err = 0;

        err = unsubscribe();
	if (err) {
		printk("Could not unsubscribe MQTT client. Error: %d\n", err);
		return -1;
	}

	err = mqtt_disconnect(&client);
	if (err) {
		printk("Could not disconnect MQTT client. Error: %d\n", err);
		return -1;
	}
}

int8_t mqtt_send(enum mqtt_qos qos, u8_t *data, size_t len)
{
        int err = 0;

        err = data_publish(&client, qos, data, len);
        if(err < 0){
                printk("fail sending message\n");
        }
}

int8_t send_one_shot_mqtt_msg(enum mqtt_qos qos, u8_t *data, size_t len) 
{
	int err = 0;

	err = mqtt_init(qos, data, len);
	if(err < 0){
		printk("fail mqtt_init\n");
	}

	err = mqtt_send(qos, data, len);
	if(err < 0){
		printk("fail send_mqtt\n");
	}

	err = mqtt_uninit(qos, data, len);
	if(err < 0){
		printk("fail mqtt_uninit\n");
	}

	return 0;
}


void modem_trace_enable()
{
    // enable modem trace
    if (at_cmd_write("AT%XMODEMTRACE=1,2", NULL, 0, NULL) != 0) {
        printk("AT%XMODEMTRACE=1,2 FAILED\n");
        return -1;
    }
    // lock in
    if (at_cmd_write("AT+CFUN=0", NULL, 0, NULL) != 0) {
        printk("AT+CFUN=0 FAILED\n");
        return -1;
    }
}


void main(void)
{
	int err;
    char buf[32];

	nrf_gnss_data_frame_t gps_data;
	u8_t		      cnt = 0;

////    nmea_strings[0] = "test\r\n";

    nmea_strings[0][0] = 'a';
    nmea_strings[0][1] = '\r';
    nmea_strings[0][2] = '\n';

    dk_leds_init();
    dk_set_led_on(0);

    // set modem to LTE-M, GPS on
    if (at_cmd_write(AT_XSYSTEMMODE, NULL, 0, NULL) != 0) {
        printk("AT_XSYSTEMMODE FAILED\n");
        return -1;
    }

//    setup_modem();

//    modem_trace_enable();
//    return 0;

    init_gps();

    client_init(&client);

    modem_configure();
    broker_init();
    mqtt_init();
    printk("The MQTT simple sample started\n");

    while (true) {
        // free socket if needed
//        if (gnss_fd != -1) {
//            printk("Releasing socket %d\n", gnss_fd);
//            nrf_close(gnss_fd);
//            gnss_fd = -1;
//        }
        // set modem to off mode: 
//        if (at_cmd_write("AT+CFUN=0", NULL, 0, NULL) != 0) {
//            printk("AT+CFUN FAILED \n");
//        } else {
//            printk("AT+CFUN SUCCEEDED \n");
//        }
//        k_sleep(K_SECONDS(1));

//        if (at_cmd_write("AT+CPSMS?", NULL, 0, NULL) != 0) {
//            printk("AT+CFUN FAILED \n");
//        } else {
//            printk("AT+CFUN SUCCEEDED \n");
//        }
//        k_sleep(K_SECONDS(1));

        if (LTEMode) {
            if (gps_active) {
                if (at_cmd_write(AT_DEACTIVATE_GPS_LTE_SAME, NULL, 0, NULL) != 0) {
                        printk("AT_DEACTIVATE_GPS_LTE_SAME FAILED\n");
                        return -1;
                }
                gps_active = false;
            }
            if (!lte_active) {
                if (at_cmd_write(AT_ACTIVATE_LTE_GPS_SAME, NULL, 0, NULL) != 0) {
                        printk("AT_ACTIVATE_LTE_GPS_SAME FAILED\n");
                        return -1;
                }
                lte_active = true;
            }
            err = mqtt_send(MQTT_QOS_1_AT_LEAST_ONCE, nmea_strings[0], strlen(nmea_strings[0]));
            if (err != 0){
                printk("Error: send_one_shot_mqtt_msg: %u\n", err);
                return -1;
            }

            connect_count++;

            printk("Connected count: %d\n", connect_count);

//            LTEMode = false;
        } else {
            // GPS get fix
//            if (init_app() != 0) {
//                return -1;
//            }
//            if (setup_modem() != 0) {
//              printk("Failed to initialize modem\n");
//              return -1;
//            }
//            if (at_cmd_write(AT_DEACTIVATE_LTE_GPS_SAME, NULL, 0, NULL) != 0) {
//                    return -1;
//            }
            if (lte_active) {
                if (at_cmd_write(AT_DEACTIVATE_LTE_GPS_SAME, NULL, 0, NULL) != 0) {
                        printk("AT_DEACTIVATE_LTE_GPS_SAME FAILED\n");
                        return -1;
                }
                lte_active = false;
            }
            if (!gps_active) {
                if (at_cmd_write(AT_ACTIVATE_GPS_LTE_SAME, NULL, 0, NULL) != 0) {
                        printk("AT_ACTIVATE_GPS_LTE_SAME FAILED\n");
                        return -1;
                }
                gps_active = true;
            }
            printk("Getting GPS data...\n");
            dk_set_led_on(1);

            while (!LTEMode) {

                do {
                    /* Loop until we don't have more
                     * data to read
                     */
                } while (process_gps_data(&gps_data) > 0);

                if (!got_first_fix) {
                    cnt++;
                    printk("\033[1;1H");
                    printk("\033[2J");
                    print_satellite_stats(&gps_data);
                    printk("\nScanning [%c] ",
                            update_indicator[cnt%4]);
                }

                if (((k_uptime_get() - fix_timestamp) >= 1) &&
                     (got_first_fix)) {
//                    LTEMode = true;
                    got_first_fix = false;
                    dk_set_led_off(0);
                    printk("\033[1;1H");
                    printk("\033[2J");

                    print_satellite_stats(&gps_data);

                    printk("---------------------------------\n");
                    print_pvt_data(&last_fix);
                    printk("\n");
                    print_nmea_data();
                    printk("---------------------------------");

                    update_terminal = false;
//                    LTEMode = true;
                }

                k_sleep(K_MSEC(500));
            }

        }
        k_sleep(K_MSEC(1000));
    }

}
